import autofit as af
from autofit_workspace.howtofit.chapter_1_introduction.tutorial_5_visualization_masking.src.dataset.dataset import (
    Dataset,
    MaskedDataset,
)
from autofit_workspace.howtofit.chapter_1_introduction.tutorial_5_visualization_masking.src.phase.result import (
    Result,
)
from autofit_workspace.howtofit.chapter_1_introduction.tutorial_5_visualization_masking.src.phase.analysis import (
    Analysis,
)


"""The Phase class ``.nit__ method is unchanged from the previous tutorial, however other methods are changed."""


class Phase(af.AbstractPhase):

    gaussian = af.PhaseProperty("gaussian")

    Result = Result

    def __init__(self, *, gaussian, search):
        """
        A phase which fits a `Gaussian` model using a `NonLinearSearch`.

        Parameters
        ----------
        gaussian : gaussians.Gaussian
            The model component `Gaussian` class fitted by this phase.
        search: class
            The class of a non_linear search
        """

        super().__init__(search=search)

        self.gaussian = gaussian

    """
    The run method is slightly different, as it now passed a mask in addition to the `Dataset`. These are used to set up
    the masked-dataset in the `analysis.py` module.
    """

    def run(self, dataset: Dataset, mask):
        """
        Pass a `Dataset` to the phase, running the phase and `NonLinearSearch`.

        Parameters
        ----------
        dataset: aa.Dataset
            The `Dataset` fitted by the phase, as defined in the `dataset.py` module.
        mask: Mask2D
            The mask used for the analysis.

        Returns
        -------
        result: AbstractPhase.Result
            A result object comprising information on the `NonLinearSearch` and the maximum likelihood model.
        """

        analysis = self.make_analysis(dataset=dataset, mask=mask)

        result = self.run_analysis(analysis=analysis)

        return self.make_result(result=result, analysis=analysis)

    def make_analysis(self, dataset, mask):
        """
        Returns an Analysis object, which creates the `Dataset` and contains the functions which perform the fit.

        Parameters
        ----------
        dataset: aa.Dataset
            The `Dataset` fitted by the phase, as defined in the `dataset.py` module.

        Returns
        -------
        analysis : Analysis
            An analysis object that the `NonLinearSearch` calls to determine the fit log_likelihood for a given model
            instance.
        """

        """To mask the `Dataset` we simply pass both to the MaskedDataset class."""

        masked_dataset = MaskedDataset(dataset=dataset, mask=mask)

        """
        The `image_path` is where visualizatiion of the model fit is output. Below, we direct it to the same path as
        the `NonLinearSearch` output, but with an additional folder `image` at the end. This path should be used
        for pretty much any project.
        """

        return Analysis(
            masked_dataset=masked_dataset, image_path=self.search.paths.image_path
        )

    def make_result(self, result, analysis):
        return self.Result(
            samples=result.samples,
            previous_model=self.model,
            search=self.search,
            analysis=analysis,
        )



import autofit as af

from autofit_workspace.howtofit.chapter_2_results.src.fit import fit as f
from autofit_workspace.howtofit.chapter_2_results.src.phase import visualizer

import pickle

# The `analysis.py` module is unchanged from the previous tutorial.


class Analysis(af.Analysis):
    def __init__(self, masked_dataset, settings, image_path=None):

        # The `MaskedDataset` and visualizer are created in the same way as tutorial 4.

        self.masked_dataset = masked_dataset
        self.settings = settings

        self.visualizer = visualizer.Visualizer(
            masked_dataset=self.masked_dataset, image_path=image_path
        )

    def log_likelihood_function(self, instance):
        """
        Returns the fit of a list of Profiles (Gaussians, Exponentials, etc.) to the dataset, using a
        model instance.

        Parameters
        ----------
        instance
            The list of Profile model instance (e.g. the Gaussians, Exponentials, etc.).

        Returns
        -------
        fit : Fit.log_likelihood
            The log likelihood value indicating how well this model fit the `MaskedDataset`.
        """
        model_data = self.model_data_from_instance(instance=instance)
        fit = self.fit_from_model_data(model_data=model_data)
        return fit.log_likelihood

    def model_data_from_instance(self, instance):
        return sum(
            [
                profile.profile_from_xvalues(xvalues=self.masked_dataset.xvalues)
                for profile in instance.profiles
            ]
        )

    def fit_from_model_data(self, model_data):
        return f.FitDataset(masked_dataset=self.masked_dataset, model_data=model_data)

    def visualize(self, instance, during_analysis):

        model_data = self.model_data_from_instance(instance=instance)
        fit = self.fit_from_model_data(model_data=model_data)

        self.visualizer.visualize_fit(fit=fit, during_analysis=during_analysis)

    def save_for_aggregator(self, paths):
        """Save files like the dataset, mask and settings as pickle files so they can be loaded in the ``Aggregator``"""

        # These functions save the objects we will later access using the aggregator. They are saved via the `pickle`
        # module in Python, which serializes the data on to the hard-disk.

        with open(f"{paths.pickle_path}/dataset.pickle", "wb") as f:
            pickle.dump(self.masked_dataset.dataset, f)

        with open(f"{paths.pickle_path}/mask.pickle", "wb") as f:
            pickle.dump(self.masked_dataset.mask, f)

        with open(f"{paths.pickle_path}/settings.pickle", "wb+") as f:
            pickle.dump(self.settings, f)



# %%
"""
Tutorial 2: Dataset Sample
==========================

In this tutorial, we'll fit multiple `Dataset`s with the same phase, producing multiple sets of results on our
hard-disk. In the following tutorials, we then use these results and the `Aggregator` to load the results into
our Jupyter notebook to interpret, inspect and plot the output results.

we'll fit 3 different `Dataset`s, each with a single `Gaussian` model.
"""

# %%
#%matplotlib inline

from pyprojroot import here

workspace_path = str(here())
#%cd $workspace_path
print(f"Working Directory has been set to `{workspace_path}`")

import autofit as af
from autofit_workspace.howtofit.chapter_2_results import src as htf
import numpy as np

# %%
"""
Here, for each `Dataset` we are going to set up the correct path, load it, create its `mask` and fit it using a `Phase`.

We want our results to be in a folder specific to the `Dataset`. we'll use the `Dataset`'s name string to do this. Lets
create a list of all 3 of our `Dataset` names.

we'll also pass these names to the `Dataset` when we create it, the name will be used by the `Aggregator` to name the 
file the data is stored. More importantly, the name will be accessible to the `Aggregator`, and we will use it to label 
figures we make via the `Aggregator`.
"""

# %%
dataset_names = ["gaussian_x1_0", "gaussian_x1_1", "gaussian_x1_2"]

# %%
"""
We can also attach information to the model-fit, by setting up an info dictionary. 

Information about our model-fit (e.g. the data of observation) that isn't part of the model-fit is made accessible to 
the `Aggregator`. For example, below we write info on the `Dataset`'s data of observation and exposure time.
"""

# %%
info = {"date_of_observation": "01-02-18", "exposure_time": 1000.0}

# %%
"""
This for loop runs over every `Dataset`, checkout the comments below for how we set up the path structure.
"""

# %%
for dataset_name in dataset_names:

    """The code below creates the `Dataset` and `mask` as per usual."""

    dataset_path = f"dataset/chapter_2/{dataset_name}"

    data = af.util.numpy_array_from_json(file_path=f"{dataset_path}/data.json")
    noise_map = af.util.numpy_array_from_json(
        file_path=f"{dataset_path}/noise_map.json"
    )
    dataset = htf.Dataset(data=data, noise_map=noise_map)

    mask = np.full(fill_value=False, shape=dataset.data.shape)

    """
    Here, we create a `Phase` as normal. However, we also include an input parameter `path_prefix`. This defines the 
    folders that the `Phase` outputs results to before the `name`. For example, if a `Phase` outputs to the path:

        `/path/to/autofit_workspace/output/name/`

    A `Phase` with the `path_prefix='phase_folder' edits this path to:

        `/path/to/autofit_workspace/output/phase_folder/name/`

    You can input multiple folders, for example `path_prefix='folder_0/folder_1' would create the path:

        `/path/to/autofit_workspace/output/folder_0/folder_1/name/`

    Below, we use the `data_name`, so our results go in a folder specific to the `Dataset`, e.g:

        `/path/to/autofit_workspace/output/gaussian_x1_0/phase_t2/`
    """

    print(
        f"Emcee has begun running - checkout the "
        f"autofit_workspace/output/howtofit/chapter_2/aggregator/{dataset_name}/phase_t2_agg folder for live "
        f"output of the results. This Jupyter notebook cell with progress once Emcee has completed - this could take a "
        f"few minutes!"
    )

    phase = htf.Phase(
        search=af.DynestyStatic(
            path_prefix=f"howtofit/chapter_2/aggregator/{dataset_name}",
            name="phase_t2_agg",
        ),
        profiles=af.CollectionPriorModel(gaussian=htf.profiles.Gaussian),
        settings=htf.SettingsPhase(),
    )

    """Note that we pass the info to the `phase` when we run it, so that the `Aggregator` can make it accessible."""

    phase.run(dataset=dataset, mask=mask, info=info)

# %%
"""
Checkout the output folder - you should see three new sets of results corresponding to our 3 `Gaussian` datasets.

Unlike previous tutorials, these folders in the output folder are named after the `Dataset` and contain the folder
with the phase`s name, as opposed to just the phase-name folder.
"""



import matplotlib.pyplot as plt

"""This function plots a line of 1D data, and was already introduce in tutorial 2."""


def line(xvalues, line, title=None, color="k", errors=None, ylabel=None):
    """Plot a 1D line of data on a plot of x versus y, where the x-axis is the x coordinate of the line and the y-axis
    is the intensity of the line at that coordinate.

    Parameters
    ----------
    xvalues : np.ndarray
        The x-coordinates the profile is defined on.
    line : np.ndarray
        The intensity values of the profile which are plotted.
    ylabel : str
        The y-label of the plot.
    """

    plt.errorbar(
        x=xvalues, y=line, yerr=errors, color=color, ecolor="k", elinewidth=1, capsize=2
    )
    plt.title(title)
    plt.xlabel("x value of profile")
    plt.ylabel(ylabel)
    plt.show()
    plt.clf()
